library nui_web;
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:nui_core/busevent/nui_busevent.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_web/nui_connection.dart';
import 'package:nui_web/nui_web_errors.dart';
import 'nui_web_models.dart';
export 'package:nui_web/nui_web.dart';
export 'package:nui_web/nui_web_models.dart';
export 'package:nui_web/nui_connection.dart';

typedef HttpHeaderSetter = Map<String, String> Function();
typedef HttpExceptionWrapper<T extends NUIEnt> = Future<NUIHttpResponse<T>> Function();
typedef HttpAdditionalHandler = void Function(NUIHttpResponse);
typedef HttpErrorHandler = void Function(BuildContext context, String code);

abstract class NUIWebInterface{

  /**
   * Download a file from a URL endpoint
   */
  Future<Uint8List?> downloadFromURL({
    required String url, //Url endpoint
  });

  /**
   * Calling a POST request to the server
   */
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /** One of the 3 body options must be provided */
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    NUIEntMapper<T>? entity, //Optional response entity mapper
    Encoding? encoding, //Optional request encoding
    bool storeTemp //Optional flag to store the response in the temp storage
  });

  /**
   * Calling a POST request to the server with multipart
   */
  Future<NUIHttpResponse<T>> doMultipartPOST<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    List<http.MultipartFile>? files, //Optional request files
    Map<String, String>? fields, //Optional request fields

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  });

  /**
   * Calling a PUT request to the server
   */
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /**One of the 3 body options must be provided **/
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  });

  /**
   * Calling a PATCH request to the server
   */
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /** One of the 3 body options must be provided **/
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  });

  /**
   * Calling a DELETE request to the server
   */
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable
    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  });

  /**
   * Calling a GET request to the server
   */
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  });
}

class NUIWeb implements NUIWebInterface{
  static const _NUIWEB_MODULE = "NUIWeb";

  static const int ERROR_INTERCEPTOR = 1090;

  /**
   * Building NUIWeb instance from the builder
   */
  NUIWeb._builder(NUIWebBuilder builder):
        _baseUrl = builder._baseUrl,
        _preHandler = builder._preHandler,
        _postHandlerDelay = builder._postHandlerDelay,
        _errorHandler = builder._errorHandler,
        _postHandler = builder._postHandler,
        _headers = builder._headers,
        _statusMessages = builder._statusMessages,
        _interceptors = builder._interceptors,
        _onRequestInterceptors = builder._onRequestInterceptors,
        _onResponseInterceptors = builder._onResponseInterceptors,
        _connectivityListener = builder._connectivityListener,
        _module = isNullOrEmpty(builder._module) ? _defaultModule : builder._module!{
    _webMap[_module] = this;
    NUISharedPrefBuilder().build();
    NUIBusEventBuilder().module(_getBusEventModule()).build();
    NUIConnectionUtil.subscribeConnectivityChange((result, hasInternet) {
      _getBusEvent().send<bool>(NUIBusEventData<bool>(data: hasInternet, sender: _getBusEventModule(), code: NUIConnectionUtil.CONNECTIVITY_CHANNEL), channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL);
      if(_connectivityListener != null) {
        _connectivityListener!(result, hasInternet);
      }
    });
  }

  final Map<int, String>? _statusMessages;
  final String _module;
  final String _baseUrl;
  final HttpAdditionalHandler? _preHandler;
  final HttpAdditionalHandler? _postHandler;
  final HttpErrorHandler? _errorHandler;
  final Duration _postHandlerDelay;
  final ConnectivityChangeListener? _connectivityListener;
  final HttpHeaderSetter? _headers;
  final List<NUIWebInterceptor> _interceptors;
  final List<NUIOnRequestInterceptor> _onRequestInterceptors;
  final List<NUIOnResponseInterceptor> _onResponseInterceptors;
  static Map<String, NUIWeb> _webMap = Map();
  static const String _defaultModule = "NUIWebDefaultModule";

  static NUIWeb get({String? module}){
    NUIWeb? instance = _webMap[isNullOrEmpty(module) ? _defaultModule : module];
    if(instance == null) throw Exception("NUIWeb -> NUIWeb instance has not been initialized with builder");
    return instance;
  }

  NUIBusEvent _getBusEvent(){
    return NUIBusEvent.get(module: _getBusEventModule());
  }

  void sendBusEvent<T>(NUIBusEventData<T> data, {String channel = NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().send<T>(data, channel: channel);
  }

  void subscribeBusEvent<T>(NUIBusEventListener listener, {String channel = NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().listen(listener, channel: channel);
  }

  void disposeBusEvent(NUIBusEventListener listener, {String channel = NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().dispose(listener, channel: channel);
  }

  String _getBusEventModule(){
    return "NUIWebBusEvent_" + (_module);
  }

  NUIWebMethod? _matchMethod(String requestMethod){
    switch(requestMethod){
      case "GET": return NUIWebMethod.GET;
      case "PUT": return NUIWebMethod.PUT;
      case "POST": return NUIWebMethod.POST;
      case "DELETE": return NUIWebMethod.DELETE;
      case "PATCH": return NUIWebMethod.PATCH;
      default: return null;
    }
  }

  /**
   * Getting the processed URL
   */
  String _getProcessedUrl(String url, String? pathVariable, Map<String, String>? params){
    String webUrl = isNullOrEmpty(_baseUrl) ? url : (_baseUrl[_baseUrl.length - 1] != '/' ? (_baseUrl + "/" + url) : (_baseUrl + url));
    if(!isNullOrEmpty(pathVariable)){
      if(webUrl[webUrl.length - 1] != '/') {
        webUrl = webUrl + "/" + pathVariable!;
      }
      else{
        webUrl = webUrl + pathVariable!;
      }
    }
    webUrl =  webUrl + _getPathParams(params);
    logNUI(NUIWeb._NUIWEB_MODULE, "URL : $webUrl");
    return webUrl;
  }

  dynamic _getContentBody(NUIEnt? bodyEnt, dynamic body, String? bodyString){
    String? bodyContent =  bodyString != null ? bodyString : (bodyEnt != null ? jsonEncode(bodyEnt) : jsonEncode(body));
    final prettyString = JsonEncoder.withIndent("   ").convert(bodyContent);
    logNUI(NUIWeb._NUIWEB_MODULE, "Request body :\n$prettyString");

    return bodyContent;
  }

  void _logWeb(String message){
    logNUI(NUIWeb._NUIWEB_MODULE, "$message");
  }

  void _preHandle<T>(NUIHttpResponse<T> response){
    if(_preHandler != null) _preHandler!(response);
  }

  void _postHandle<T>(NUIHttpResponse<T> response){
    if(_postHandler != null){
      NUIAsync.delay(_postHandlerDelay, () {
        _postHandler!(response);
      });
    }
  }

  /**
   * Getting path parameters
   */
  String _getPathParams(Map<String, String>? params){
    if(params == null || params.isEmpty){
      return "";
    }
    String pathParam = "";
    for(String key in params.keys){
      pathParam = pathParam + (pathParam.isEmpty ? "?$key=${params[key]}" : "&$key=${params[key]}");
    }
    return pathParam;
  }

  /**
   * Getting headers
   */
  Map<String, String>? _getHeaders(Map<String, String>? additionalHeaders){
    if(_headers == null) {
      _logWeb("Headers : " + additionalHeaders.toString());
      return additionalHeaders;
    }
    if(additionalHeaders == null) {
      _logWeb("Headers : " + _headers!().toString());
      return _headers!();
    }

    Map<String, String> headers = _headers!();
    for(String key in additionalHeaders.keys){
      headers[key] = additionalHeaders[key]!;
    }
    if(headers.isNotEmpty){
      _logWeb("Headers : " + headers.toString());
    }
    return headers;
  }

  Future<NUIHttpResponse<T>> wrapException<T extends NUIEnt>(NUIEntMapper<T>? entity, HttpExceptionWrapper<T> task) async{
    try {
      return task();
    }catch(e){
      _logWeb("Encountered an exception : ${e.toString()}");
      return NUIHttpResponse(success: false, code: HttpStatus.internalServerError, message: e.toString(), body: null);
    }
  }

  Future<bool> _checkInternetConnection() async{
    return NUIConnectionUtil.hasConnection();
  }

  void _storeCacheResponse({required String url, String? pathVariable, required http.Response response}){
    if(response.statusCode != HttpStatus.ok) return;
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    final cacheResponse = NUIHttpCacheResponse(response: response.body, headers: response.headers);
    NUISharedPref.get().save(fullUrl, cacheResponse);
  }

  NUIHttpCacheResponse? getCacheResponse({required String url, String? pathVariable}){
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    return NUISharedPref.get().quickGetObject(fullUrl, NUIHttpCacheResponseEntity());
  }

  /**
   * Processing the response
   */
  Future<NUIHttpResponse<D>> _processResponse<D>(http.Response response, NUIEntMapper<D>? entity) async {
    final interceptedResponse = await processOnResponseIntercept(responseBody: response.body, entity: entity);
    String? finalBody = interceptedResponse.noIntercepting ? response.body : interceptedResponse.interceptedResponseBody;

    D? entityValue;
    List<D>? listEntityValue;

    _logWeb("Response body : $finalBody");

    /**
     * If the entity class is provided, parse the response into the entity class
     */
    if(entity != null) {
      try {
        final bodyMap = jsonDecode(finalBody);
        if (bodyMap is List) {
          listEntityValue = asListOfObjects(entity, bodyMap);
          entityValue = entity.toEntity({});
        }
        else {
          entityValue = entity.toEntity(bodyMap);
        }
      } catch (e) {
        entityValue = entity.toEntity({});
        logNUI(_NUIWEB_MODULE,
            "Failed to decode stream response with error: ${e.toString()}");
      }
    }

    if (response.statusCode == HttpStatus.ok) {
      return NUIHttpResponse<D>(
          success: true,
          code: response.statusCode,
          message: _statusMessages?[response.statusCode] ?? "Http status OK",
          body: entityValue,
          listBody: listEntityValue,
          responseBody: jsonDecode(finalBody),
          response: finalBody,
          headers: response.headers
      );
    }
    else {
      _processError(response.statusCode.toString());
      // Map<String, dynamic>? decodedBody;
      // try{
      //   decodedBody = jsonDecode(finalBody);
      // }catch(e){
      //   logNUI("NUIWeb", "Failed to decode failure response with error : $e");
      //   //Error message might be the response body as string itself
      // }

      return NUIHttpResponse<D>(
          success: false,
          code: response.statusCode,
          message: _statusMessages?[response.statusCode] ?? "No message specified",
          body: entityValue,
          listBody: listEntityValue,
          responseBody: jsonDecode(finalBody),
          response: finalBody,
          headers: response.headers
      );
    }
  }

  Future<NUIHttpResponse<D>> _processStreamResponse<D>(http.StreamedResponse response, NUIEntMapper<D>? entity, String respBody) async{
    // final resp = respBody;
    D? entityValue;
    List<D>? listEntityValue;

    if(entity != null) {
      try {
        final bodyMap = jsonDecode(respBody);
        if (bodyMap is List) {
          listEntityValue = asListOfObjects(entity, bodyMap);
          entityValue = entity.toEntity({});
        }
        else {
          entityValue = entity.toEntity(bodyMap);
        }
      } catch (e) {
        entityValue = entity.toEntity({});
        logNUI(_NUIWEB_MODULE,
            "Failed to decode stream response with error: ${e.toString()}");
      }
    }

    if(response.statusCode == HttpStatus.ok){
      return NUIHttpResponse<D>(
          success: true,
          code: response.statusCode,
          body: entityValue,
          listBody: listEntityValue,
          message: _statusMessages?[response.statusCode] ?? "Http status OK",
          response: respBody,
          headers: response.headers
      );
    }
    else {
      _processError(response.statusCode.toString());
      String? errorMessage;

      return NUIHttpResponse<D>(
          success: false,
          code: response.statusCode,
          body: entityValue,
          listBody: listEntityValue,
          message: _statusMessages?[response.statusCode] ?? errorMessage ?? "No message specified",
          response: respBody,
          headers: response.headers
      );
    }
  }

  void _processError(String code){
    if(_errorHandler != null){
      final context = NUINavigator.getLastContext();
      if(context != null && _errorHandler != null) {
        _errorHandler!(context, code);
      }
    }
  }

  /**
   * Pre api request checking
   */
  Future<NUIHttpResponse<T>?> _preRequestChecking<T extends NUIEnt>(String url, NUIWebMethod method, Map<String, String>? headers, dynamic body, NUIEntMapper<T>? entity) async{
    final hasConnection = await _checkInternetConnection();
    if(hasConnection) {
      //Has connection, now run interceptors
      final interceptResult = await _processInterceptors(url, method, headers, body);
      if(interceptResult != null){
        return NUIHttpResponse<T>(success: interceptResult.success, code: interceptResult.code ?? NUIWeb.ERROR_INTERCEPTOR, body: null, message: interceptResult.message, responseBody: interceptResult.data);
      }
      return null;
    }
    else{
      _processError(NUIWebErrors.NO_INTERNET_CONNECTION.code.toString());
      return NUIHttpResponse<T>(success: false, code: NUIWebErrors.NO_INTERNET_CONNECTION.code, body: null, message: "No internet connection found");
    }
  }

  /**
   * Processing list of interceptors
   */
  Future<NUIStandardHttpResponse?> _processInterceptors(String url, NUIWebMethod method, Map<String, String>? headers, dynamic body) async{
    //No interceptors, so return true
    if(isNullOrEmpty(_interceptors)) return null;

    for(NUIWebInterceptor interceptor in _interceptors){
      final result = await interceptor.intercept(url, method, dynamic, headers, body);
      if(result.interceptResponse == true) return result;
    }
    return null;
  }

  Future<NUIInterceptor> processOnRequestIntercept({required String url, String? pathVariables, Map<String, String>? params, Map<String, String> headers = const {}, dynamic body, NUIEnt<dynamic>? bodyEnt, String? bodyString}) async {

    NUIInterceptor finalInterceptedRequest = NUIInterceptor(
        interceptedUrl: url,
        interceptedPathVariables: pathVariables,
        interceptedHeaders: headers,
        interceptedParams: params,
        interceptedBody: body,
        interceptedBodyEnt: bodyEnt,
        interceptedBodyString: bodyString
    );
    if(_onRequestInterceptors.isNullOrEmpty()){
      return finalInterceptedRequest;
    }
    for(NUIOnRequestInterceptor interceptor in _onRequestInterceptors){
      finalInterceptedRequest.noIntercepting = false;
      final interceptedRequest = await interceptor.onRequestInterceptor(finalInterceptedRequest.interceptedUrl, finalInterceptedRequest.interceptedPathVariables, finalInterceptedRequest.interceptedParams, finalInterceptedRequest.interceptedHeaders, finalInterceptedRequest.interceptedBody, finalInterceptedRequest.interceptedBodyEnt, finalInterceptedRequest.interceptedBodyString);

      finalInterceptedRequest.interceptedUrl = interceptedRequest.interceptedUrl;
      finalInterceptedRequest.interceptedParams = interceptedRequest.interceptedParams ?? finalInterceptedRequest.interceptedParams;
      finalInterceptedRequest.interceptedPathVariables = interceptedRequest.interceptedPathVariables ?? finalInterceptedRequest.interceptedPathVariables;
      finalInterceptedRequest.interceptedBody = interceptedRequest.interceptedBody ?? finalInterceptedRequest.interceptedBody;
      finalInterceptedRequest.interceptedHeaders = interceptedRequest.interceptedHeaders ?? finalInterceptedRequest.interceptedHeaders;
      finalInterceptedRequest.interceptedBodyEnt = interceptedRequest.interceptedBodyEnt ?? finalInterceptedRequest.interceptedBodyEnt;
      finalInterceptedRequest.interceptedBodyString = interceptedRequest.interceptedBodyString ?? finalInterceptedRequest.interceptedBodyString;
      if(interceptedRequest.stopIntercepting){
        return finalInterceptedRequest;
      }
    }
    return finalInterceptedRequest;
  }

  Future<NUIInterceptor> processOnResponseIntercept({required String responseBody, NUIEntMapper? entity}) async {
    NUIInterceptor finalInterceptedResponse = NUIInterceptor.forResponse(interceptedResponseBody: responseBody);
    if(_onResponseInterceptors.isNullOrEmpty()){
      return finalInterceptedResponse;
    }
    for(NUIOnResponseInterceptor interceptor in _onResponseInterceptors){
      finalInterceptedResponse.noIntercepting = false;
      final interceptedResponse = await interceptor.onResponseInterceptor(finalInterceptedResponse.interceptedResponseBody);

      finalInterceptedResponse.interceptedResponseBody = interceptedResponse.interceptedResponseBody;
      if(interceptedResponse.stopIntercepting){
        return finalInterceptedResponse;
      }

    }
    return finalInterceptedResponse;
  }

  @override
  Future<Uint8List?> downloadFromURL({required String url}) async{
    try {
      http.Client client = new http.Client();
      var req = await client.get(Uri.parse(url));
      var bytes = req.bodyBytes;
      return bytes;
    }catch(e){
      return null;
    }
  }

  @override
  Future<NUIHttpResponse<T>> doMultipartPOST<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    List<http.MultipartFile>? files, //Optional request files
    Map<String, String>? fields, //Optional request fields

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.POST, headers, fields, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header

    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: null, bodyEnt: null, bodyString: null, headers: headers, params: params, pathVariables: pathVariable);
    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);
    // final targetUrl = _getProcessedUrl(url, pathVariable, params);
    var request = http.MultipartRequest("POST", Uri.parse(targetUrl));
    request.fields.addAll(fields ?? {});
    request.files.addAll(files ?? []);
    final allHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers) : _getHeaders(interceptedRequest.interceptedHeaders);
    if(!isNullOrEmpty(allHeaders)){
      for(var key in allHeaders!.keys){
        request.headers[key] = allHeaders[key]!;
      }
    }

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.StreamedResponse response = await request.send();
      final resp = await (response.stream.bytesToString());
      print("NUIWeb: Response body : " + resp);
      return _processStreamResponse<T>(response, entity, resp);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /** One of the 3 body options must be provided */
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    NUIEntMapper<T>? entity, //Optional response entity mapper
    Encoding? encoding, //Optional request encoding
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.POST, headers, body, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header
    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: body, bodyEnt: bodyEnt, bodyString: bodyString, headers: headers, params: params, pathVariables: pathVariable);
    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);

    final contentBody = interceptedRequest.noIntercepting ? _getContentBody(bodyEnt, body, bodyString)
        : _getContentBody(interceptedRequest.interceptedBodyEnt, interceptedRequest.interceptedBody, interceptedRequest.interceptedBodyString);

    final processedHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers)
        : _getHeaders(interceptedRequest.interceptedHeaders);

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.Response response = await http.post(
          Uri.parse(targetUrl),
          headers: processedHeaders,
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /**One of the 3 body options must be provided **/
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.PUT, headers, body, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header
    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: body, bodyEnt: bodyEnt, bodyString: bodyString, headers: headers, params: params, pathVariables: pathVariable);
    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);

    final contentBody = interceptedRequest.noIntercepting ? _getContentBody(bodyEnt, body, bodyString)
        : _getContentBody(interceptedRequest.interceptedBodyEnt, interceptedRequest.interceptedBody, interceptedRequest.interceptedBodyString);

    final processedHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers)
        : _getHeaders(interceptedRequest.interceptedHeaders);

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.Response response = await http.put(
          Uri.parse(targetUrl),
          headers: processedHeaders,
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable

    /** One of the 3 body options must be provided **/
    NUIEnt? bodyEnt, //Optional request body entity
    dynamic body, //Optional request body
    String? bodyString, //Optional request body string

    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.PATCH, headers, body, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header
    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: body, bodyEnt: bodyEnt, bodyString: bodyString, headers: headers, params: params, pathVariables: pathVariable);
    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);

    final contentBody = interceptedRequest.noIntercepting ? _getContentBody(bodyEnt, body, bodyString)
        : _getContentBody(interceptedRequest.interceptedBodyEnt, interceptedRequest.interceptedBody, interceptedRequest.interceptedBodyString);

    final processedHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers)
        : _getHeaders(interceptedRequest.interceptedHeaders);

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.Response response = await http.patch(
          Uri.parse(targetUrl),
          headers: processedHeaders,
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable
    Encoding? encoding, //Optional request encoding
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.DELETE, headers, null, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header
    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: null, bodyEnt: null, bodyString: null, headers: headers, params: params, pathVariables: pathVariable);

    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);

    final processedHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers)
        : _getHeaders(interceptedRequest.interceptedHeaders);

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.Response response = await http.delete(
          Uri.parse(targetUrl),
          headers: processedHeaders
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({
    required String url, //Url endpoint
    Map<String, String>? headers, //Optional request headers
    Map<String, String>? params, //Optional request parameters
    String? pathVariable, //Optional path variable
    NUIEntMapper<T>? entity, //Optional response entity mapper
    bool storeTemp = false //Optional flag to store the response in the temp storage
  }) async{

    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.GET, headers, null, entity);
    if(preprocess != null) return preprocess;
    //Add the initial headers into API header
    if(headers == null){headers = <String, String>{};}
    headers.addAll(_headers != null ? _headers!() : <String, String>{});
    final interceptedRequest = await processOnRequestIntercept(url: url, body: null, bodyEnt: null, bodyString: null, headers: headers, params: params, pathVariables: pathVariable);

    final targetUrl = interceptedRequest.noIntercepting ? _getProcessedUrl(url, pathVariable, params)
        : _getProcessedUrl(interceptedRequest.interceptedUrl, interceptedRequest.interceptedPathVariables, interceptedRequest.interceptedParams);

    final processedHeaders = interceptedRequest.noIntercepting ? _getHeaders(headers)
        : _getHeaders(interceptedRequest.interceptedHeaders);

    NUIHttpResponse<T> response = await wrapException<T>(entity, () async{
      http.Response response = await http.get(
        Uri.parse(targetUrl),
        headers: processedHeaders,
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }
}

class NUIWebBuilder{
  Map<int, String>? _statusMessages;
  late String _baseUrl;
  HttpHeaderSetter? _headers;
  String? _module;
  HttpAdditionalHandler? _preHandler;
  HttpAdditionalHandler? _postHandler;
  Duration _postHandlerDelay = const Duration(milliseconds: 500);
  ConnectivityChangeListener? _connectivityListener;
  HttpErrorHandler? _errorHandler;
  //Direct Interceptor
  List<NUIWebInterceptor> _interceptors = [];
  //OnRequest/OnResponse Interceptor
  List<NUIOnRequestInterceptor> _onRequestInterceptors = [];
  List<NUIOnResponseInterceptor> _onResponseInterceptors = [];

  NUIWebBuilder module(String module){
    _module = module;
    return this;
  }
  NUIWebBuilder errorHandler(HttpErrorHandler errorHandler){
    _errorHandler = errorHandler;
    return this;
  }
  NUIWebBuilder preHandler(HttpAdditionalHandler handler){
    _preHandler = handler;
    return this;
  }
  NUIWebBuilder addInterceptor(NUIWebInterceptor interceptor){
    _interceptors.add(interceptor);
    return this;
  }
  NUIWebBuilder addOnRequestInterceptors(List<NUIOnRequestInterceptor> onRequestInterceptors){
    _onRequestInterceptors.addAll(onRequestInterceptors);
    return this;
  }
  NUIWebBuilder addOnResponseInterceptors(List<NUIOnResponseInterceptor> onResponseInterceptors){
    _onResponseInterceptors.addAll(onResponseInterceptors);
    return this;
  }
  NUIWebBuilder postHandler(HttpAdditionalHandler handler, {Duration delay = const Duration(milliseconds: 500)}){
    _postHandler = handler;
    _postHandlerDelay = delay;
    return this;
  }
  NUIWebBuilder connectivityListener(ConnectivityChangeListener listener){
    this._connectivityListener = listener;
    return this;
  }
  NUIWebBuilder statusMessages(Map<int, String> messages){
    _statusMessages = messages;
    return this;
  }
  NUIWebBuilder baseURL(String url){
    _baseUrl = url;
    return this;
  }
  NUIWebBuilder headers(HttpHeaderSetter headers){
    _headers = headers;
    return this;
  }
  Future<NUIWeb> build() async{
    final web = NUIWeb._builder(this);
    logNUI(NUIWeb._NUIWEB_MODULE, "Initialization successful");
    return web;
  }
}
